from numpy import argsort as numpy_argsort
from numpy import dtype   as numpy_dtype

from operator  import attrgetter, itemgetter
from itertools import izip

from .ancillaryvariables import AncillaryVariables
from .comparison         import gt
from .coordinate         import AuxiliaryCoordinate
from .fieldlist          import FieldList
from .transform          import Transform
from .units              import Units
from .functions          import (flat, RTOL, ATOL, equals, SET_FREE_MEMORY,
                                 hash_array)

from .data.data import Data

dtype_float = numpy_dtype(float)

# --------------------------------------------------------------------
# Global properties, as defined in Appendix A of the CF conventions.
# --------------------------------------------------------------------
_global_properties = set(('comment',
                          'Conventions',
                          'history',
                          'institution',
                          'references',
                          'source',
                          'title',
                          ))

# --------------------------------------------------------------------
# Data variable properties, as defined in Appendix A of the CF
# conventions, without those which are not simple. And less
# 'long_name'.
# --------------------------------------------------------------------
_data_properties = set(('add_offset',
                        'cell_methods',
                        '_FillValue',
                        'flag_masks',
                        'flag_meanings',
                        'flag_values',
                        'missing_value',
                        'scale_factor',
                        'standard_error_multiplier',
                        'standard_name',
                        'units',
                        'valid_max',
                        'valid_min',
                        'valid_range',
                        ))

_no_units = Units()

class _Meta(object):
    '''

A summary of a field.

This object contains everything you need to know in order to aggregate
the field.

'''
    #
    _canonical_units = {}

    #
    _canonical_cell_methods = []

    def __init__(self, f, rtol=None, atol=None, messages=False,
                 no_strict_units=False, no_strict_identity=False,
                 equal_all=False, equal_ignore=(), equal=(),
                 exist_all=False, exist_ignore=(), exist=(),
                 dimension=()):
        '''

**initialization**

:Parameters:

    f : Field

    messages : bool, optional
        See the `aggregate` function for details.

    no_strict_units : bool, optional
        See the `aggregate` function for details.

    no_strict_identity : bool, optional
        See the `aggregate` function for details.

    rtol : float, optional
        See the `aggregate` function for details.

    atol : float, optional
        See the `aggregate` function for details.
   
    dimension : (sequence of) str, optional
        See the `aggregate` function for details.

**Examples**

'''
        self._nonzero = False

        self.messages = messages
        
        # ------------------------------------------------------------
        # Field
        # ------------------------------------------------------------
        self.field    = f
        self._hasData = f._hasData
        self.identity = f.identity()

        if self.identity is None:
            if not no_strict_identity and self._hasData:
                if messages:
                    print(
"Field has data but no identity and no_strict_identity=%s: %s" %
(no_strict_identity, repr(f)))
                return
        else:
            if not self._hasData:
                if messages:
                    print(
"Field has identity but no data and no_strict_identity=%s: %s" %
(no_strict_identity, repr(f)))
                return
        #--- End: if

        domain = f.domain
        items  = domain.items
 
        # ------------------------------------------------------------
        # Promote selected properties to 1-d, size 1 auxiliary
        # coordinates
        # ------------------------------------------------------------
        for property in dimension:
            value = f.getprop(property, None)
            if value is None:
                continue

            aux_coord = AuxiliaryCoordinate(properties={'long_name': property},
                                            attributes={'id': property},
                                            data=Data([value], units=''),
                                            copy=False)
            axis = domain.new_axis_identifier()
            domain.insert_aux(aux_coord, axes=[axis], copy=False)

            f.delprop(property) ### dch COPY issue?
        #--- End: for

        self.units = self.canonical_units(f, self.identity,
                                          no_strict_units=no_strict_units)

        self.cell_methods = self.canonical_cell_methods(rtol=rtol, atol=atol)

        # ------------------------------------------------------------
        # Formula_terms
        # ------------------------------------------------------------
        transforms = items(role='t')
        if not transforms:
            self.transforms    = ()
            self.formula_terms = ()
            self.grid_mapping  = ()
        else:
            transforms = transforms.values()
            self.transforms = sorted(transforms, key=attrgetter('name'))

            self.formula_terms = []
            self.grid_mapping  = []            
            for transform in transforms:
                if transform.grid_mapping:
                    self.grid_mapping.append(transform.name)
                else:
                    self.formula_terms.append(transform.name)
            #--- End: for
        #--- End: if

        # ------------------------------------------------------------
        # Ancillary variables
        # ------------------------------------------------------------
        self.ancillary_variables = self.find_ancillary_variables()
        if self.ancillary_variables is None:            
            return
           
        # ------------------------------------------------------------
        # Coordinate and cell measure arrays
        # ------------------------------------------------------------
        self.hash_values  = {}
        self.first_values = {}
        self.last_values  = {}
        self.first_bounds = {}
        self.last_bounds  = {}

        # Map axis canonical identities to their domain identifiers
        # For example: {'time': 'dim2'}
        self.id_to_axis = {}
        
        # Map domain axis identifiers to their canonical identities
        # For example: {'dim2': 'time'}
        self.axis_to_id = {}

        # Dictionaries mapping domain auxiliary coordinate identifiers
        # to their auxiliary coordiante objects
        aux_1d = items(role='a', rank=1)
            
        # A set containing the identity of each domain coordinate
        # For exanmple: set(['time', 'height', 'latitude', 'longitude'])
        self.all_coord_identities = set()

     
        # ----------------------------------------------------------------
        # 1-d coordinates
        # ----------------------------------------------------------------
        self.axis = {}
        for axis in domain.dimension_sizes:
    
            # List some information about each 1-d coordinate which
            # spans this axis. The order of elements is arbitrary, as
            # ultimately it will get sorted by each element's 'name'
            # key values.
            #
            # For example:
            # >>> info
            # [{'name': 'time', 'key': 'dim0', 'units': <CF Units: ...>},
            #  {'name': 'forecast_ref_time', 'key': 'aux0', 'units': <CF Units: ...>}]
            dim_info = []

            dim_coord = domain.item(axis)

            if dim_coord is not None:
                # This domain axis has a dimension coordinate

                identity = self.coord_has_identity_and_data(dim_coord)
                if identity is None:
                    return

                # Find the canonical units for this dimension
                # coordinate
                units = self.canonical_units(dim_coord, identity,
                                             no_strict_units=no_strict_units)
    
                dim_info.append(
                    {'identity'  : identity,
                     'key'       : axis,
                     'units'     : units,
                     'isbounded' : dim_coord._isbounded,
                     'transforms': self.find_transforms(axis, dim_coord)})

#                dim_coord_id = identity
#    
#            else:
#                # This axis does not have a dimension coordinate
#                dim_coord_id = None
            #--- End: if
    
            # Find the 1-d auxiliary coordinates which span this axis
            aux_coords = {}
            for aux in aux_1d.keys():
                if axis in domain.item_axes(aux): #dimensions[aux]:
                    aux_coords[aux] = aux_1d.pop(aux)
            #--- End: for
    
            aux_info = []
            for aux, aux_coord in aux_coords.iteritems():
                
                identity = self.coord_has_identity_and_data(aux_coord)
                if identity is None:
                    return
    
                # Find the canonical units for this 1-d auxiliary
                # coordinate
                units = self.canonical_units(aux_coord, identity,
                                             no_strict_units=no_strict_units)

                aux_info.append(
                    {'identity'  : identity,
                     'key'       : aux,
                     'units'     : units,
                     'isbounded' : aux_coord._isbounded,
                     'transforms': self.find_transforms(aux, aux_coord)})
            #--- End: for
    
            # Sort the 1-d auxiliary coordinate information
            aux_info.sort(key=itemgetter('identity'))
    
            # Prepend the dimension coordinate information to the
            # auxiliary coordinate information
            info = dim_info + aux_info
            if not info:
                if messages:
                    print(
"Can't aggregate %s: An axis has no 1-d coordinates" % repr(f))
                return
            #--- End: if

            # Find the canonical identity for this axis
            identity = info[0]['identity']
    
            self.axis[identity] = \
                {'ids'       : tuple([i['identity']   for i in info]),
                 'keys'      : tuple([i['key']        for i in info]),
                 'units'     : tuple([i['units']      for i in info]),
                 'isbounded' : tuple([i['isbounded']  for i in info]),
                 'transforms': tuple([i['transforms'] for i in info])}
            
#            d = self.axis[identity]
#            if dim_coord_id is not None:
            if dim_info:
                self.axis[identity]['dim_coord_index'] = 0 #d['ids'].index(dim_coord_id)
            else:
                self.axis[identity]['dim_coord_index'] = None
    
            self.id_to_axis[identity] = axis
            self.axis_to_id[axis]     = identity
        #--- End: for
    
        # Create a sorted list of the axes' canonical identities
        # For example: ['latitude', 'longitude', 'time']
        self.axis_ids = sorted(self.axis)
    
#        self.calculate_hash_values = set(self.axis_ids) ## dch ??

        # ------------------------------------------------------------
        # N-d auxiliary coordinates
        # ------------------------------------------------------------
        self.nd_aux = {}
        for aux, nd_aux_coord in items(role='a', rank=gt(1)).iteritems():
           
            # Find this N-d auxiliary coordinate's identity
            identity = self.coord_has_identity_and_data(nd_aux_coord)
            if identity is None:
                return

            # Find the canonical units
            units = self.canonical_units(nd_aux_coord, identity,
                                         no_strict_units=no_strict_units)
            
            # Find axes' canonical identities
            axes = [self.axis_to_id[axis] for axis in domain.item_axes(aux)]
            axes = tuple(sorted(axes))

            self.nd_aux[identity] = {
                'key'       : aux,
                'units'     : units,
                'axes'      : axes,
                'isbounded' : nd_aux_coord._isbounded,
                'transforms': self.find_transforms(aux, nd_aux_coord)}
        #--- End: for
    
        # ------------------------------------------------------------
        # Cell measures
        # ------------------------------------------------------------
        self.cm = {}
        info    = {}
        for key, cm in items(role='c').iteritems():
            
            if not self.cell_measure_has_data_and_units(cm):
                return

            # Find the canonical units for this cell measure
            units = self.canonical_units(cm, cm.identity(),
                                         no_strict_units=no_strict_units)
            
            # Find axes' canonical identities
            axes = [self.axis_to_id[axis] for axis in domain.item_axes(key)]
            axes = tuple(sorted(axes))
            
            if units in info:
                # Check for ambiguous cell measures, i.e. those which
                # have the same units and span the same axes.
                for value in info[units]:
                    if axes == value['axes']:
                        if messages:
                           print(
"Can't aggregate %s: Ambiguous %s cell measure objects" %
(repr(f), repr(units)))
                        return
            else:
                info[units] = []
            #--- End: if
    
            info[units].append({'key' : key,
                                'axes': axes})
        #--- End: for
    
        # For each cell measure's canonical units, sort the
        # information by axis identities.
        for units, value in info.iteritems():
            value.sort(key=itemgetter('axes'))        
            self.cm[units] = {'keys': tuple([v['key']  for v in value]),
                              'axes': tuple([v['axes'] for v in value])}
        #--- End: for

        # ------------------------------------------------------------
        # Properties and attributes
        # ------------------------------------------------------------
        properties = set(f.properties)
        properties.difference_update(_data_properties)

        # Properties which must exist in both fields and are equal
        x = []

        if equal:            
            p = dict([(prop, f.getprop(prop)) for prop in equal
                      if prop not in equal_ignore and prop in properties])
            properties.difference_update(p)
            x.append(tuple(sorted(p.items())))
        else:
            x.append(None)

        if equal_all:        
            p = dict([(prop, f.getprop(prop))
                      for prop in (properties.difference(('long_name',)) - 
                                   _global_properties.union(exist_ignore))])
            properties.difference_update(p)
            x.append(tuple(sorted(p.items())))
        else:
            x.append(None)

        self.equal_properties = tuple(x)

        # Properties which might not exist in either field and are not
        # necessarily equal
        self.properties = properties        

        # Properties which must exist in both fields but are not
        # necessarily equal
        if exist_all:
            exist_properties = properties.difference(('long_name',))
            exist_properties.difference_update(_global_properties)
        else:
            exist_properties = set()

        if exist:
            exist_properties.update(exist)

        exist_properties.difference_update(exist_ignore)

        self.exist_properties = tuple(sorted(exist_properties))

        # Attributes
        self.attributes = set(('file',))

        # ----------------------------------------------------------------
        # Still here? Then create the structural signature.
        # ----------------------------------------------------------------
        self.structural_signature()

        # Initialize the flag which indicates whether or not this
        # field has already been aggregated
        self.aggregated_field = False

        self.sort_indices = {}
        self.sort_keys    = {}
  
        self._nonzero = True
    #--- End: def

    def __nonzero__(self):
        '''
x.__nonzero__() <==> bool(x)

'''
        return self._nonzero
    #--- End: if

    def __repr__(self):
        '''
x.__repr__() <==> repr(x)

'''
        return '<CF %s: %s>' % (self.__class__.__name__,
                                repr(getattr(self, 'field', None)))
    #--- End: def

    def __str__(self):
        '''
x.__str__() <==> str(x)

'''
        strings = []
        for attr in sorted(self.__dict__):
            strings.append('%s.%s = %s' % (self.__class__.__name__, attr,
                                           repr(getattr(self, attr))))
            
        return '\n'.join(strings)
    #--- End: def

    def canonical_units(self, variable, identity, no_strict_units=False):
        '''

Updates the `_canonical_units` attribute.

:Parameters:

    variable : cf.Variable

    identity : str

    no_strict_units : bool 
        See the `cf.aggregate` for details.

:Returns:

    out : cf.Units or None

**Examples**

'''
        var_units = variable.Units

        _canonical_units = self._canonical_units

        if identity in _canonical_units:
            if var_units:
                for u in _canonical_units[identity]:
                    if var_units.equivalent(u):
                        return u
                #--- End: for
    
                # Still here?
                _canonical_units[identity].append(var_units)

            elif no_strict_units or variable.dtype.kind == 'S':
                var_units = _no_units
        else:
            if var_units:
                _canonical_units[identity] = [var_units]                
            elif no_strict_units or variable.dtype.kind == 'S':
                var_units = _no_units
        #--- End: if

        # Still here?
        return var_units
    #--- End: def

    def canonical_cell_methods(self, rtol=None, atol=None):
        '''

Updates the `_canonical_cell_methods` attribute.

:Parameters:

    atol : float

    rtol : float

:Returns:

    out : cf.CellMethods or None

**Examples**

'''
        cell_methods = getattr(self.field, 'cell_methods', None)

        if cell_methods is None:
            return None

        _canonical_cell_methods = self._canonical_cell_methods

        for cm in _canonical_cell_methods:
            if cell_methods.equivalent(cm, rtol=rtol, atol=atol):
                return cm
        #--- End: for
               
        # Still here?
        _canonical_cell_methods.append(cell_methods)

        return cell_methods
    #--- End: def

    def cell_measure_has_data_and_units(self, cm):
        '''

:Parameters:

    cm : cf.CellMethods

:Returns:

    out : bool

**Examples**

'''
        if cm.Units and cm._hasData:
            return True
        
        if self.messages:
            print("Cell measure object has no units and/or no data: %s" %
                  repr(self.field))
    #--- End: def

    def coord_has_identity_and_data(self, coord):
        '''

:Parameters:

    coord : cf.Coordinate

:Returns:

    out : str or None
        The coordinate object's identity, or None if there is no
        identity and/or no data.

**Examples**

'''
        identity = coord.identity()

        if identity is not None:
            all_coord_identities = self.all_coord_identities

            if identity in all_coord_identities:
                if self.messages:
                    print("Field has more than one %s coordinate objects: %s" %
                          (repr(identity), repr(self.field)))
                return None
            #--- End: if

            if coord._hasData or (coord._isbounded and coord.bounds._hasData):
                all_coord_identities.add(identity)
                return identity
        #--- End: if

        if self.messages:
            print("Coordinate object has no identity and/or no data: %s" %
                  repr(self.field))
            
        # Return None
        return None
    #--- End: def

    def find_ancillary_variables(self):
        '''

:Returns:

    out : dict or None

**Examples**

'''
        f = self.field

        f_ancillary_variables = getattr(f, 'ancillary_variables', None)

        if f_ancillary_variables is None:
            return {}

        ancillary_variables = {}
        for av in f_ancillary_variables:
            identity = av.identity()
            if identity in ancillary_variables:
                if self.messages:
                    print("Field has more than one %s ancillary variable: %s" %
                          (repr(identity), repr(f)))

                return None
            #--- End: if
            ancillary_variables[identity] = av
        #--- End: for

        return ancillary_variables
    #--- End: def

    def structural_signature(self, global_properties=None,
                             ignore_properties=None):
        '''

:Returns:

    out : tuple

**Examples**

'''
        f = self.field    
      
        # Initialize the structual signature the presence of a data
        # array, the identity, the canonical units and the canonical
        # cell methods.
        signature = [self._hasData, self.identity, self.units, self.cell_methods]

        # Add properties
        signature_append = signature.append

        signature_append(self.equal_properties)
        
        signature_append(self.exist_properties)

        # Add Flags
        signature_append(getattr(f, 'Flags', None))
        
        # Add valid_min, valid_max, valid_range,
        # standard_error_multiplier
        signature.extend((f.getprop('valid_min'                , None),
                          f.getprop('valid_max'                , None),
                          f.getprop('valid_range'              , None),
                          f.getprop('standard_error_multiplier', None)))

        # Add ancillary variables
        if self.ancillary_variables:
            signature_append(tuple(sorted(self.ancillary_variables)))
        else:
            signature_append(None)
        
        # Add transforms
        signature_append(tuple(self.grid_mapping))
        signature_append(tuple(self.formula_terms))

        # Add 1-d coordinates
        axis = self.axis        
        x = [(axis[identity]['ids'],
              axis[identity]['units'],
              axis[identity]['isbounded'],
              axis[identity]['transforms']) for identity in self.axis_ids]
        signature_append(tuple(x))
        
        x = [False if axis[identity]['dim_coord_index'] is None else True
             for identity in self.axis_ids]
        signature_append(tuple(x))
        
        # Add N-d auxiliary coordinates
        nd_aux = self.nd_aux
        x = [(identity,
              nd_aux[identity]['units'],
              nd_aux[identity]['axes'],
              nd_aux[identity]['isbounded'],
              nd_aux[identity]['transforms']) for identity in sorted(nd_aux)]
        signature_append(tuple(x))
        
        # Add cell measures
        cm = self.cm
        x = [(units,
              cm[units]['axes']) for units in sorted(cm)]
        signature_append(tuple(x))
        
        self.signature = tuple(signature)
    #--- End: def

    def find_transforms(self, key, coord):
        '''

:Parameters:

    key : str
        The domain identifier of the coordinate object

    coord : cf.Coordinate

:Returns:

    out : tuple or None

**Examples**

>>> dim_coord
<CF DimensionCoordinate: ....>
>>> m.find_transforms('dim0', dim_coord)

>>> aux_coord
<CF AuxiliaryCoordinate: ....>
>>> m.find_transforms('aux1', aux_coord)

'''    
        transforms = self.transforms

        if not transforms:
            return None

        # Select the transforms which contain coordinate pointers
        transforms = [t.name for t in transforms if key in t.coordinates]
        
        if not transforms:
            return None

        return tuple(transforms)
    #--- End: def

#--- End: class

def aggregate(fields, messages=False, no_strict_units=False, no_overlap=False,
              contiguous=False, _delete_meta=True, no_strict_identity=False,
              equal_all=False, equal_ignore=(), equal=(),
              exist_all=False, exist_ignore=(), exist=(),
              dimension=(), concatenate=True, copy=True, 
              axes=None, donotchecknonaggregatingaxes=False,
              debug=0):
    '''

Aggregate fields into as few fields as possible.

The aggregation of fields may be thought of as the combination fields
into each other to create a new field that occupies a larger domain.

Using the CF aggregation rules, input fields are separated into
aggregatable groups and each group (which may contain just one field)
is then aggregated to a single field. These aggregated fields are
returned in a field list.

:Parameters:

    fields : (sequence of) Field or FieldList
        The field or fields to aggregated.

    messages : bool, optional
        If True then print messages giving reasons why particular
        fields have not been aggregated.

    no_overlap : bool, optional
        If True then require that aggregated fields have adjacent
        dimension coordinate object cells which do not overlap (but
        they may share common boundary values). Ignored if the
        dimension coordinates objects do not have bounds. See the
        *contiguous* parameter.

    contiguous : bool, optional
        If True then require that aggregated fields have adjacent
        dimension coordinate object cells which partially overlap or
        share common boundary values. Ignored if the dimension
        coordinate objects do not have bounds. See the *no_overlap*
        parameter.

    no_strict_units : bool, optional
        If True then assume that fields or domain items (such as
        coordinate objects) with the same identity (as returned by
        their `!identity` methods) but missing units all have
        equivalent but unspecified units, so that aggregation may
        occur. By default such fields are not aggregatable.

    no_strict_identity : bool, optional
        If True then treat fields with data arrays but with no
        identities (as returned by their `~cf.Field.identity` methods)
        as having equal but unspecified identities, so that
        aggregation may occur. By default such fields are not
        aggregatable.

    equal_all : bool, optional
        If True then require that aggregated fields have the same set
        of non-standard CF properties (including
        `~cf.Field.long_name`), with the same values. See the
        *concatenate* parameter.

    equal_ignore : (sequence of) str, optional
        Specify CF properties to omit from any properties specified by
        or implied by the *equal_all* and *equal* parameters.

    equal : (sequence of) str, optional
        Specify CF properties for which it is required that aggregated
        fields all contain the properties, with the same values. See
        the *concatenate* parameter.

    exist_all : bool, optional
        If True then require that aggregated fields have the same set
        of non-standard CF properties (including, in this case,
        long_name), but not requiring the values to be the same. See
        the *concatenate* parameter.

    exist_ignore : (sequence of) str, optional
        Specify CF properties to omit from the properties specified by
        or implied by the *exist_all* and *exist* parameters.

    exist : (sequence of) str, optional
        Specify CF properties for which it is required that aggregated
        fields all contain the properties, but not requiring the
        values to be the same. See the *concatenate* parameter.

    dimension : (sequence of) str, optional
        Create new axes for each input field which has one or more of
        the given properties. For each CF property name specified, if
        an input field has the property then, prior to aggregation, a
        new axis is created with an auxiliary coordinate whose datum
        is the property's value and the property itself is deleted
        from that field.

    concatenate : bool, optional
        If False then a CF property is omitted from an aggregated
        field if the property has unequal values across constituent
        fields or is missing from at least one constituent field. By
        default a CF property in an aggregated field is the
        concatenated collection of the distinct values from the
        constituent fields, delimited with the string
        ``' :AGGREGATED: '``.

    copy : bool, optional
        If False then do not copy fields prior to aggregation.
        Setting this option to False may change input fields in place,
        and the output fields may not be independent of the
        inputs. However, if it is known that the input fields are
        never to accessed again (such as in this case: ``f =
        cf.aggregate(f)``) then setting *copy* to False can reduce the
        time taken for aggregation.

    axes : (sequence of) str, optional
        Select axes to aggregate over. Aggregation will only occur
        over as large a subset as possible of these axes. Each axis is
        identified by the exact identity of a one dimensional
        coordinate object, as returned by its `!identity`
        method. Aggregations over more than one axis will occur in the
        order given. By default, aggregation will be over as many axes
        as possible.

    donotchecknonaggregatingaxes : bool, optional
        If True, and *axes* is set, then checks for consistent data
        array values will only be made for one dimensional coordinate
        objects which span the any of the given aggregating axes. This
        can reduce the time taken for aggregation, but if any those
        checks would have failed then this clearly allows the
        possibility of an incorrect result. Therefore, this option
        should only be used in cases for which it is known that the
        non-aggregating axes are in fact already entirely consistent.

    debug : int, optional
        Print debugging information for each input field. If 1 then
        print each field's structural signature. If 2 the print the
        entire metadata summary for each field, which includes the
        structural signature. By default no debugging information is
        printed.

:Returns:

    out : cf.FieldList
        The aggregated fields.
    
**Examples**

The following six fields comprise eastward wind at two different times
and for three different atmospheric heights for each time:

>>> f
[<CF Field: eastward_wind(latitude(73), longitude(96)>,
 <CF Field: eastward_wind(latitude(73), longitude(96)>,
 <CF Field: eastward_wind(latitude(73), longitude(96)>,
 <CF Field: eastward_wind(latitude(73), longitude(96)>,
 <CF Field: eastward_wind(latitude(73), longitude(96)>,
 <CF Field: eastward_wind(latitude(73), longitude(96)>]
>>> g = cf.aggregate(f)
>>> g
[<CF Field: eastward_wind(height(3), time(2), latitude(73), longitude(96)>]
>>> g[0].source
'Model A'
>>> g = cf.aggregate(f, dimension=('source',))
[<CF Field: eastward_wind(source(1), height(3), time(2), latitude(73), longitude(96)>]
>>> g[0].source
AttributeError: 'Field' object has no attribute 'source'

'''

    SET_FREE_MEMORY()

    output_fields = FieldList()

    output_fields_append = output_fields.append

    atol = ATOL()
    rtol = RTOL()

    if axes is not None and isinstance(axes, basestring):
        axes = (axes,)

    # Parse parameters
    properties = {'equal'       : equal,
                  'exist'       : exist, 
                  'equal_ignore': equal_ignore,
                  'exist_ignore': exist_ignore}

    for key, value in properties.iteritems():
        if not value:
            continue

        if isinstance(equal, basestring):
            # If it is a string then convert to a single element
            # sequence
            properties[key] = (value,)
        else:
            try:
                value[0]
            except TypeError:
                raise TypeError("Bad type of '%s' parameter: %s" % (key, value))
    #--- End: for

    equal        = properties['equal']
    exist        = properties['exist']
    equal_ignore = properties['equal_ignore']
    exist_ignore = properties['exist_ignore']

    # ----------------------------------------------------------------
    # Group fields with the same structural signature
    # ----------------------------------------------------------------
    signatures = {}
    for f in flat(fields):

        # Create the metadata summary, including the structural
        # signature
        meta = _Meta(f, messages=messages, rtol=rtol, atol=atol,
                     no_strict_units=no_strict_units, 
                     no_strict_identity=no_strict_identity,
                     equal_all=equal_all,
                     equal_ignore=equal_ignore,
                     equal=equal,
                     exist_all=exist_all,
                     exist_ignore=exist_ignore,
                     exist=exist,
                     dimension=dimension)

        if not meta:
            # Can't find a structural signature for this field, so it
            # can't be aggregated. Put it straight into the output
            # list and move on to the next input field.
            if not copy:
                output_fields_append(f)
            else:
                output_fields_append(f.copy())

            continue
        #--- End: if

        # Append this field to the list of fields with the same
        # structural signature
        signatures.setdefault(meta.signature, []).append(meta)

        if debug:
            if hasattr(meta.field, 'file'):
                print 'FILE:', meta.field.file
            else:
                print 'NO FILE'

            if debug == 1:
                print 'STRUCTURAL SIGNATURE:', meta.signature, '\n'
            elif debug == 2:
                print meta, '\n'
        #--- End: if
    #--- End: for    

    # ----------------------------------------------------------------
    # Within each group of fields with the same structural signature,
    # aggregate as many fields as possible. Sort the signatures so
    # that independent aggregations of the same set of input fields
    # return fields in the same order.
    # ----------------------------------------------------------------
    for signature in sorted(signatures):

        meta = signatures[signature]

        if len(meta) == 1:
            # There's only one field with this signature, so we can
            # add it straight to the output list and move on to the
            # next signature.
            if not copy:       
                output_fields_append(meta[0].field) 
            else:
                output_fields_append(meta[0].field.copy()) 

            if messages:
                print("Field has unique structural signature: %s" %
                      repr(meta[0].field))

            continue
        #--- End: if

        # ------------------------------------------------------------
        # Still here? Then there are 2 or more fields with this
        # signature which may be aggregatable. These fields need to be
        # passed through until no more aggregations are possible. With
        # each pass, the number of fields in the group will reduce by
        # one for each aggregation that occurs. Each pass represents
        # an aggregation in another axis.
        # ------------------------------------------------------------

        # ------------------------------------------------------------
        # For each axis's 1-d coordinates, create the canonical hash
        # value and the first and last datums.
        # ------------------------------------------------------------
        if axes is None: 
            aggregating_axes = meta[0].axis_ids
            _create_hash_and_first_values(meta, None, False)
        else:            
            aggregating_axes = []
            axis_items = meta[0].axis.items()
            for axis in axes:
                coord = meta[0].field.coord(axis, exact=True)
                if coord is None:
                    continue

                coord_identity = coord.identity()
                for identity, value in axis_items:
                    if (identity not in aggregating_axes and 
                        coord_identity in value['ids']):
                        aggregating_axes.append(identity)                        
                        break
            #--- End: for
            _create_hash_and_first_values(meta, aggregating_axes, 
                                          donotchecknonaggregatingaxes)
        #--- End: if

        for axis in aggregating_axes:
#        while 1:
            number_of_fields = len(meta)
            if number_of_fields == 1:
                break

#            # --------------------------------------------------------
#            # For each axis's 1-d coordinates, create the canonical
#            # hash value and first (and last) datum.
#            # --------------------------------------------------------
#            _create_hash_and_first_values(meta)

            # --------------------------------------------------------
            # Separate the fields with the same structural signature
            # into groups such that either within each group the
            # fields' domains differ in only one axis or each group
            # contains only one field. Note that the 'a_identity'
            # attribute is set in the _group_fields function.
            # --------------------------------------------------------
            grouped_meta = _group_fields(meta, axis)

            if len(grouped_meta) == number_of_fields:
                continue

            # --------------------------------------------------------
            # Within each group, aggregate as many fields as possible.
            # --------------------------------------------------------
            for m in grouped_meta:

                if len(m) == 1:
                    continue

                # ----------------------------------------------------
                # Still here? The sort the fields in place by the
                # canonical first values of their 1-d coordinates for
                # the aggregating axis.
                # ----------------------------------------------------
                _sorted_by_first_values(m, axis)

                # ----------------------------------------------------
                # Check that the aggregating axis's 1-d coordinates
                # don't overlap, and don't aggregate anything in this
                # group if any do.
                # ----------------------------------------------------
                if not _ok_coordinate_arrays(m, axis, no_overlap, contiguous):
                    continue

                # ----------------------------------------------------
                # Still here? Then pass through the fields
                # ----------------------------------------------------
                while len(m) >= 2:
                    new = _aggregate_2_fields(m[0], m[1],
                                              rtol=rtol, atol=atol,
                                              contiguous=contiguous,
                                              no_overlap=no_overlap,
                                              no_strict_units=no_strict_units,
                                              messages=messages,
                                              concatenate=concatenate,
                                              copy=copy)

                    if new:
                        # Successfully aggregated these two fields, so
                        # replace the two fields in m with the their
                        # aggregate.
                        m[0:2] = [new] 
                    else:
                        # Couldn't aggregate these two fields, so
                        # remove the first one from m an put it into
                        # the list of output fields.
                        if not copy:       
                            output_fields_append(m.pop(0).field)
                        else:
                            output_fields_append(m.pop(0).field.copy())
                #--- End: while

            #--- End: for

            # Re-assemble the fields as a single list
            meta = [m for gm in grouped_meta for m in gm]

#            if len(meta) == number_of_fields:
#                # ----------------------------------------------------
#                # No aggregation took place on this pass, so we're
#                # done and can add these fields to the output list.
#                # ----------------------------------------------------
#                break            
#        #--- End: while
        #--- End: for

        output_fields.extend((m.field for m in meta))
    #--- End: for

    SET_FREE_MEMORY(None)

    return output_fields
#--- End: def

def _create_hash_and_first_values(meta, axes, donotchecknonaggregatingaxes):
    '''

Updates each field's _Meta object.

:Parameters:

    meta : list of _Meta

:Returns:

    None

'''
    for m in meta:
        domain = m.field.domain
        domain_dimensions = domain.dimensions

        m_sort_keys    = m.sort_keys
        m_sort_indices = m.sort_indices

        m_hash_values  = m.hash_values
        m_first_values = m.first_values
        m_last_values  = m.last_values

        m_id_to_axis = m.id_to_axis
        # --------------------------------------------------------
        # Create a hash value for each metadata array
        # --------------------------------------------------------
        
        # --------------------------------------------------------
        # 1-d coordinates
        # --------------------------------------------------------
        for identity in m.axis_ids:
            
            if (axes is not None and donotchecknonaggregatingaxes and
                identity not in axes):
                x = [None] * len(m.axis[identity]['keys'])
                m_hash_values[identity]  = x
                m_first_values[identity] = x[:]
                m_last_values[identity]  = x[:]
                continue

            # Still here?
            m_axis_identity = m.axis[identity]

            axis = m_id_to_axis[identity]

            dim_coord = domain.get(axis, None)

            # Find the sort indices for this axis ...
            if dim_coord is not None:
                # ... which has a dimension coordinate
                m_sort_keys[axis] = axis
                if not domain.direction(axis):
                    # Axis is decreasing
                    sort_indices = slice(None, None, -1)
                    null_sort = False
                else:
                    # Axis is increasing
                    sort_indices = slice(None)
                    null_sort = True
             
            else:
                # ... which doesn't have a dimension coordinate but
                #     does have one or more 1-d auxiliary coordinates
                aux = m_axis_identity['keys'][0]
                sort_indices = numpy_argsort(domain.get(aux).unsafe_array)
                m_sort_keys[axis] = aux 
                null_sort = False
            #-- End: if
            m_sort_indices[axis] = sort_indices

            hash_values  = []
            first_values = []
            last_values  = []
#            first_bounds = []
#            last_bounds  = []

            for key, canonical_units in izip(m_axis_identity['keys'],
                                             m_axis_identity['units']):

                coord = domain.get(key)
                coord_units = coord.Units

                # Change the coordinate data type if required
#                if coord.dtype not in (dtype_float, dtype_str):
                if coord.dtype.char not in ('d', 'S'):
                    coord = coord.copy(_only_Data=True)
                    coord.dtype = dtype_float

                # Change the coordinate's units to the canonical ones
                coord.Units = canonical_units

                # Get the coordinate's data array
                if null_sort:
                    array = coord.Data.unsafe_array
                else:
                    array = coord.Data.array[sort_indices]

                hash_value = hash_array(array)

                first_values.append(array.item(0)) #[0])
                last_values.append(array.item(-1)) #[-1])

                if coord._isbounded:
                    if null_sort:
                        array = coord.bounds.Data.unsafe_array
                    else:
                        array = coord.bounds.Data.array[sort_indices, ...]

                    hash_value = (hash_value, hash_array(array))

                    if key[:3] == 'dim':  # can do better than this!
                        # Record the bounds of the first and last
                        # (sorted) cells of a dimension coordinate
                        # (don't need to do this for an auxiliary
                        # coordinate).
                        array0 = array[0, ...]
                        array0.sort()
                        m.first_bounds[identity] = array0

                        array0 = array[-1, ...]
                        array0.sort()
                        m.last_bounds[identity] = array0
                #--- End: if
                    
                hash_values.append(hash_value)

                # Reinstate the coordinate's original units
                coord.Units = coord_units
            #--- End: for
                
            m_hash_values[identity]  = hash_values
            m_first_values[identity] = first_values
            m_last_values[identity]  = last_values
        #--- End: for

        # ------------------------------------------------------------
        # N-d auxiliary coordinates
        # ------------------------------------------------------------
        if donotchecknonaggregatingaxes:
            for aux in m.nd_aux.itervalues():
                aux['hash_value'] = None
        else:
            for aux in m.nd_aux.itervalues():
                      
                key = aux['key']
                
                sort_indices = [m_sort_indices[dim] 
                                for dim in domain_dimensions[key]]
                
                coord = domain.get(key)
            
                coord_units = coord.Units
                
                # Change the coordinate data type if required
#                if coord.dtype not in (dtype_float, dtype_str):
                if coord.dtype.char not in ('d', 'S'):
                    coord = coord.copy(_only_Data=True)
                    coord.dtype = dtype_float
                    copied = True
                else:
                    copied = False
                    
                # Change the coordinate's units to the canonical ones
                coord.Units = aux['units'] #canonical_units
                
                axes = [m_id_to_axis[identity] for identity in aux['axes']]
                if axes != coord.Data._dimensions:
                    if not copied:
                        coord = coord.copy(_only_Data=True)
            
                    coord.transpose(axes)
                #--- End: if
            
                # Get the coordinate's data array
                array = coord.Data.array[tuple(sort_indices)]
                
                hash_value = hash_array(array)
            
                if coord._isbounded:
                    sort_indices.append(Ellipsis)
                    array = coord.bounds.Data.array[tuple(sort_indices)]
                    hash_value = (hash_value, hash_array(array))
                    
                aux['hash_value'] = hash_value
                
                # Reinstate the coordinate's original units
                coord.Units = coord_units
            #--- End: for
        #--- End: if
            
        # ------------------------------------------------------------
        # Cell measures
        # ------------------------------------------------------------
        if donotchecknonaggregatingaxes:
            for cm in m.cm.itervalues():            
                cm['hash_values'] = [None] * len(cm['keys'])
        else:

            for units, cm in m.cm.iteritems():
                
                hash_values = []
#                for key, axes, hash_value in izip(cm['keys'],
#                                                      cm['axes'],
#                                                      cm['hash_values']):
            
                for key, axes in izip(cm['keys'], cm['axes']):
            
                    sort_indices = [m_sort_indices[dim]
                                    for dim in domain_dimensions[key]]
            
                    coord = domain.get(key)
                    coord_units = coord.Units
                    
                    # Change the coordinate data type if required
#                    if coord.dtype not in (dtype_float, dtype_str):
                    if coord.dtype.char not in ('d', 'S'):
                        coord = coord.copy(_only_Data=True)
                        coord.dtype = dtype_float
                        copied = True
                    else:
                        copied = False
                        
                    # Change the coordinate's units to the canonical ones
                    coord.Units = units
                    
                    axes = [m_id_to_axis[identity] for identity in axes]
                    if axes != coord.Data._dimensions:
                        if not copied:
                            coord = coord.copy(_only_Data=True)
                            
                        coord.transpose(axes)
                    #--- End: if
                    
                    array = coord.Data.array[tuple(sort_indices)]
                
                    hash_values.append(hash_array(array)) 
            
                    # Reinstate the coordinate's original units
                    coord.Units = coord_units  
                #--- End: for
            
                cm['hash_values'] = hash_values
            #--- End: for
        #--- End: if

#        m.calculate_hash_values = set()
    #--- End: for
#--- End: def

def _group_fields(meta, axis):
    '''

:Parameters:

    meta : list of _Meta

    axis : str
        The name of the axis to group for aggregation.

:Returns:

    out : list of FieldList

'''
    axes = meta[0].axis_ids

    if axes:
        if axis in axes:
            axes = axes[:]
            axes.remove(axis)

        if axes:
            sort_by_axis_ids = itemgetter(*axes)
            
            def _hash_values(m):
                return sort_by_axis_ids(m.hash_values)

            meta.sort(key=_hash_values)
        #--- End: if
    #--- End: if

    # Create a new group of potentially aggregatable fields (which
    # contains the first field in the sorted list)
    m0 = meta[0]
    groups_of_fields = [[m0]]

    hash0 = m0.hash_values

    for m0, m1 in izip(meta[:-1], meta[1:]):

        #-------------------------------------------------------------
        # Count the number of axes which are different between the two
        # fields
        # -------------------------------------------------------------
        count = 0
        hash1 = m1.hash_values
        for identity, value in hash0.iteritems():
            if value != hash1[identity]:
                count += 1
                a_identity = identity                
        #--- End: for
        hash0 = hash1

        if count == 1:
            # --------------------------------------------------------
            # Exactly one axis has different 1-d coordinate values
            # --------------------------------------------------------
            if a_identity != axis:
                # ... but it's not the axis that we're trying
                # currently to aggregate over.
                groups_of_fields.append([m1])
                continue

            # Still here? Then it is the axis that we're trying
            # currently to aggregate over.
            ok = True

            # Check the N-d auxiliary coordinates
            for identity, aux0 in m0.nd_aux.iteritems():
                if (a_identity not in aux0['axes'] and 
                    aux0['hash_value'] != m1.nd_aux[identity]['hash_value']):
                    # This matching pair of N-d auxiliary coordinates
                    # does not span the aggregating axis and they have
                    # different data array values
                    ok = False
                    break
            #--- End: for

            if not ok:
                groups_of_fields.append([m1])
                continue 
                
            # Still here? Then check the cell measures
            cm0 = m0.cm
            for units in cm0:
                for axes, hash_value0, hash_value1 in izip(
                    cm0[units]['axes'],
                    cm0[units]['hash_values'],
                    m1.cm[units]['hash_values']):
                    
                    if a_identity not in axes and hash_value0 != hash_value1:
                        # There is a matching pair of cell measures
                        # with these units which does not span the
                        # aggregating axis and they have different
                        # data array values
                        ok = False
                        break
            #--- End: for

            if not ok:
                groups_of_fields.append([m1])
                continue 

            # Still here? Then set the identity of the aggregating
            # axis
            m0.a_identity = a_identity
            m1.a_identity = a_identity
            
            # Append field1 to this group of potentially aggregatable
            # fields
            groups_of_fields[-1].append(m1)

        else:
            # --------------------------------------------------------
            # Zero or 2 or more axes has different 1-d coordinate
            # values
            # --------------------------------------------------------
            # Create a new group of potentially aggregatable fields
            # (which contains field1).
            groups_of_fields.append([m1])

            if not count:   
                # No axes have different 1-d coordinate values
                if m0.messages:
                    print(
"%s fields have identical 1-d coordinate object values" %
repr(m0.identity))
        #--- End: if
    #--- End: for

    return groups_of_fields
#--- End: def

def _sorted_by_first_values(meta, axis):
    '''

Sort fields inplace

:Parameters:

    meta : list of _Meta

    axis : str

:Returns:

    None

''' 
    sort_by_axis_ids = itemgetter(axis)

    def _first_values(m):
        return sort_by_axis_ids(m.first_values)
    #--- End: def

    meta.sort(key=_first_values)
#--- End: def

def _ok_coordinate_arrays(meta, axis, no_overlap, contiguous):
    '''

Return True if the aggregating axis's 1-d coordinates are all
aggregatable.

It is assumed that the input metadata objects have already been sorted
by the canonical first values of their 1-d coordinates.

:Parameters:

    meta : list of _Meta

    axis : str
        Find the canonical identity of the aggregating axis.

    no_overlap : bool
        See the `aggregate` function for details.

    contiguous : bool
        See the `aggregate` function for details.

:Returns:

    out : bool

**Examples**

>>> if not _ok_coordinate_arrays(meta, True, False)
...     print "Don't aggregate"

'''
    m = meta[0]

    dim_coord_index = m.axis[axis]['dim_coord_index']

    if dim_coord_index is not None:
        # ------------------------------------------------------------
        # The aggregating axis has a dimension coordinate
        # ------------------------------------------------------------
        # Check for overlapping dimension coordinate cell centres
        dim_coord_index0 = dim_coord_index

        for m0, m1 in izip(meta[:-1], meta[1:]):
            dim_coord_index1 = m1.axis[axis]['dim_coord_index']
            if (m0.last_values[axis][dim_coord_index0] >=
                m1.first_values[axis][dim_coord_index1]):
                # Found overlap
                if m.messages:
                    print(
"Won't aggregate group containing %s: Overlapping %s coordinate values (%s >= %s)" %
(repr(m.field),
 repr(m.axis[axis]['ids'][dim_coord_index]),
 m0.last_values[axis][dim_coord_index0],
 m1.first_values[axis][dim_coord_index1]))
                return

            dim_coord_index0 = dim_coord_index1        
        #--- End: for

        if axis in m.first_bounds:
            # --------------------------------------------------------
            # The dimension coordinates have bounds
            # --------------------------------------------------------
            if no_overlap:
                for m0, m1 in izip(meta[:-1], meta[1:]):
                    if (m1.first_bounds[axis][0] <
                        m0.last_bounds[axis][1]):
                        # Do not aggregate anything in this group
                        # because overlapping has been disallowed and
                        # the first cell from field1 overlaps with the
                        # last cell from field0.
                        if m.messages:
                            print(
"Won't aggregate group containing %s: no_overlap=True and overlapping %s coordinate bounds values (%s < %s)" %
(repr(m.field),
repr(m.axis[axis]['ids'][dim_coord_index]),
 m1.first_bounds[axis][0],
 m0.last_bounds[axis][1]
 ))
                        return
                #--- End: for

            else:
                for m0, m1 in izip(meta[:-1], meta[1:]):
                    m0_last_bounds  = m0.last_bounds[axis]        
                    m1_first_bounds = m1.first_bounds[axis]
                    if (m1_first_bounds[0] <= m0_last_bounds[0] or
                        m1_first_bounds[1] <= m0_last_bounds[1]):
                        # Do not aggregate anything in this group
                        # because, even though overlapping has been
                        # allowed, the first cell from field1 overlaps
                        # in an unreasonable way with the last cell
                        # from field0.
                        if m.messages:
                            print(
"Can't aggregate group containing %s: Unreasonably overlapping %s coordinate bounds values (%s <= %s or %s <= %s)" %
(repr(m.field),
 repr(m.axis[axis]['ids'][dim_coord_index]),
 m1_first_bounds[0], m0_last_bounds[0],
 m1_first_bounds[1], m0_last_bounds[1]
 ))
                        return
                #--- End: for
            #--- End: if

            if contiguous:
                for m0, m1 in izip(meta[:-1], meta[1:]):
                    if (m0.last_bounds[axis][1] <
                        m1.first_bounds[axis][0]):
                        # Do not aggregate anything in this group
                        # because contiguous coordinates have been
                        # specified and the first cell from field1 is
                        # not contiguous with the last cell from
                        # field0.
                        if m.messages:
                            print(
"Won't aggregate group containing %s: contiguous=True and noncontiguous %s coordinate bounds values (%s < %s)" %
(repr(m.field),
 repr(m.axis[axis]['ids'][dim_coord_index]),
 m0.last_bounds[axis][1], 
 m1.first_bounds[axis][0]
 ))
                        return
                #--- End: for
            #--- End: if
        #--- End: if

    else:
        # ------------------------------------------------------------
        # The aggregating axis does not have a dimension coordinate,
        # but it does have at least one 1-d auxiliary coordinate.
        # ------------------------------------------------------------
        # Check for duplicate auxiliary coordinate values
        for i, identity in enumerate(meta[0].axis[axis]['ids']):
            set_of_1d_aux_coord_values    = set()
            number_of_1d_aux_coord_values = 0
            for m in meta:
                aux = m.axis[axis]['keys'][i]
                array = m.field.domain.get(aux).array
                set_of_1d_aux_coord_values.update(array)
                number_of_1d_aux_coord_values += array.size
                if len(set_of_1d_aux_coord_values) != number_of_1d_aux_coord_values:
                    # Found duplicate auxiliary coordinate values
                    if m.messages:
                        print(
"Won't aggregate group containing %s: Duplicate %s auxiliary coordinate values" %
(repr(m.field), repr(identity)))
                        return
            #--- End: for
        #--- End: for
    #--- End: if
 
    # ----------------------------------------------------------------
    # Still here? Then the aggregating axis does not overlap between
    # any of the fields.
    # ----------------------------------------------------------------
    return True
#--- End: def

def _aggregate_2_fields(m0, m1, rtol=None, atol=None, messages=False,    
                        no_strict_units=False, no_overlap=False, 
                        contiguous=False, concatenate=True, copy=True):
    '''

:Parameters:

    m0 : _Meta

    m1 : _Meta

    contiguous : bool, optional
        See the `aggregate` function for details.
   
    rtol : float, optional
        See the `aggregate` function for details.

    atol : float, optional
        See the `aggregate` function for details.
   
    messages : bool, optional
        See the `aggregate` function for details.
   
    no_overlap : bool, optional
        See the `aggregate` function for details.
  
    no_strict_units : bool, optional
        See the `aggregate` function for details.

:Returns:

    out : _Meta or bool
  
''' 
    if copy and not m0.aggregated_field:
        m0.field = m0.field.copy()

    a_identity = m0.a_identity
    
    # ----------------------------------------------------------------
    # Aggregate transforms
    # ----------------------------------------------------------------
    if m0.formula_terms:
        t = _aggregate_transforms(m0, m1, rtol=rtol, atol=atol,
                                  no_strict_units=no_strict_units,
                                  no_overlap=no_overlap, messages=messages,
                                  contiguous=contiguous)
        if not t:
            return
    else:
        t = None
  
    # ----------------------------------------------------------------
    # Aggregate ancillary variables
    # ----------------------------------------------------------------
    if m0.ancillary_variables:
        av = _aggregate_ancillary_variables(m0, m1, rtol=rtol,
                                            atol=atol,
                                            no_strict_units=no_strict_units,
                                            no_overlap=no_overlap,
                                            messages=messages,
                                            contiguous=contiguous)
        if not av:
            return
    else:
        av = None
 
    # Still here?
    field0 = m0.field
    field1 = m1.field
    if copy:
        field1 = field1.copy()

    domain0 = field0.domain
    domain1 = field1.domain

    if t:
        # ------------------------------------------------------------
        # Update transforms
        # ------------------------------------------------------------
        for key, transform in t.iteritems():
            domain0.insert_transform(transform, key=key, copy=False, replace=True)
    #--- End: if

    if av:
        # ------------------------------------------------------------
        # Update ancillary variables
        # ------------------------------------------------------------
        field0.ancillary_variables = av

    # ----------------------------------------------------------------
    # Map the axes of field1 to those of field0
    # ----------------------------------------------------------------
    dim1_name_map = {'bounds': 'bounds'}
    for identity in m0.axis_ids:
        dim1_name_map[m1.id_to_axis[identity]] = m0.id_to_axis[identity]

    # ----------------------------------------------------------------
    # In each field, find the identifier of the aggregating axis.
    # ----------------------------------------------------------------
    adim0 = m0.id_to_axis[a_identity]
    adim1 = m1.id_to_axis[a_identity]

    # ----------------------------------------------------------------
    # In each field, direction of the aggregating axis.
    # ----------------------------------------------------------------
    adim_directions0 = domain0.direction(adim0)
    adim_directions1 = domain1.direction(adim1)

    # ----------------------------------------------------------------
    # 
    # ----------------------------------------------------------------
#    pmsize1 = domain1.get(m1.sort_keys[adim1]).Data.pmsize
    if adim_directions0:
#        pmsize0 = domain0.get(m0.sort_keys[adim0]).Data.pmsize
        # The aggregating axis in field0 is increasing ...
        push = True
        if not adim_directions1:
            # ... but the aggregating axis in field1 is decreasing
            reverse = True
        else:
            # ... and the aggregating axis in field1 is also
            # increasing
            reverse = False
    else:
        # The aggregating axis in field0 is decreasing ...
        push = False
        if not adim_directions1:
            # ... and the aggregating axis in field1 is also
            # decreasing
            reverse = False
        else:
            # ... but the aggregating axis in field1 is increasing
            reverse = True
    #--- End: if
            
    # ----------------------------------------------------------------
    # Find matching pairs of coordinates and cell measures which span
    # the aggregating axis
    # ----------------------------------------------------------------
    # 1-d coordinates
    spanning_variables = [(domain0.get(key0), domain1.get(key1))
                          for key0, key1 in izip(m0.axis[a_identity]['keys'],
                                                 m1.axis[a_identity]['keys'])] 
   
    hash_values0 = m0.hash_values[a_identity]
    hash_values1 = m1.hash_values[a_identity]
    for i, (hash0, hash1) in enumerate(izip(hash_values0, hash_values1)):
        try:
            hash_values0[i].append(hash_values1[i])
        except AttributeError:
            hash_values0[i] = [hash_values0[i], hash_values1[i]]
    #--- End: for

    # N-d auxiliary coordinates
    for identity in m0.nd_aux:
        aux0 = m0.nd_aux[identity]
        aux1 = m1.nd_aux[identity]
        if a_identity in aux0['axes']:
            spanning_variables.append((domain0.get(aux0['key']),
                                       domain1.get(aux1['key'])))

            hash_value0 = aux0['hash_value']
            hash_value1 = aux1['hash_value']
            try:
                hash_value0.append(hash_value1)
            except AttributeError:
                aux0['hash_value'] = [hash_value0, hash_value1]
    #--- End: for
    
    # Cell measures                
    for units in m0.cm:
        hash_values0 = m0.cm[units]['hash_values']
        hash_values1 = m1.cm[units]['hash_values']
        for i, (axes, key0, key1) in enumerate(izip(m0.cm[units]['axes'],
                                                    m0.cm[units]['keys'],
                                                    m1.cm[units]['keys'])):
            if a_identity in axes:
                spanning_variables.append((domain0.get(key0), domain1.get(key1)))

                try:
                    hash_values0[i].append(hash_values1[i])
                except AttributeError:
                    hash_values0[i] = [hash_values0[i], hash_values1[i]]
    #--- End: for

    # ----------------------------------------------------------------
    # For each matching pair of coordinates and cell measures which
    # span the aggregating axis, insert the one from field1 into the
    # one from field0
    # ----------------------------------------------------------------
    for variable0, variable1 in spanning_variables:
        variable0._insert_data(variable1, push, reverse,
                               adim0, adim_directions0, dim1_name_map)
    #--- End: for        
        
    # ----------------------------------------------------------------
    # Insert the data array from field1 into the data array of field0
    # ----------------------------------------------------------------
    if m0._hasData:
        field0._insert_data(field1, push, reverse,
                            adim0, adim_directions0, dim1_name_map)
        
        # Update the axes of the data array in field0        
        domain0.dimensions['data'] = field0.Data._dimensions[:]        
    #--- End: if

    # Update the size of the Aggregating Axis in field0
    domain0.dimension_sizes[adim0] += domain1.dimension_sizes[adim1]

    # Make sure that field0 has a standard_name, if possible.
    if getattr(field0, 'id', None) is not None:
        standard_name = field1.getprop('standard_name', None)
        if standard_name is not None:
            field0.standard_name = standard_name
            del field0.id
    #--- End: if

    #-----------------------------------------------------------------
    # Update the properties in field0
    #-----------------------------------------------------------------
    for prop in set(field0._simple_properties()) | set(field1._simple_properties()):
        value0 = field0.getprop(prop, None)
        value1 = field1.getprop(prop, None)
        if equals(value0, value1):
            continue

        if concatenate:
            if value1 is not None:
                if value0 is not None:
                    field0.setprop(prop, '%s :AGGREGATED: %s' % (value0, value1))
                else:
                    field0.setprop(prop, ' :AGGREGATED: %s' % value1)
        else:
            m0.properties.discard(prop)
            if value0 is not None:
                field0.delprop(prop)
    #--- End: for

    #-----------------------------------------------------------------
    # Update the attributes in field0
    #-----------------------------------------------------------------
    for attr in m0.attributes | m1.attributes:
        value0 = getattr(field0, attr, None)
        value1 = getattr(field1, attr, None)
        if equals(value0, value1):
            continue

        if concatenate:
            if value1 is not None:
                if value0 is not None:
                    setattr(field0, attr, '%s :AGGREGATED: %s' % (value0, value1))
                else:
                    setattr(field0, attr, ' :AGGREGATED: %s' % value1)
        else:
            m0.attributes.discard(attr)
            if value0 is not None:
                delattr(field0, attr)
    #--- End: for

    # Note that the field in this _Meta object has already been
    # aggregated
    m0.aggregated_field = True

    # ----------------------------------------------------------------
    # Return the _Meta object containing the aggregated field
    # ----------------------------------------------------------------
    return m0
#--- End: def

def _aggregate_transforms(m0, m1, rtol=None, atol=None, 
                          no_strict_units=False, no_overlap=False,
                          messages=False, contiguous=False):
    '''

Aggregate fields in transforms.

:Parameters:

    m0 : _Meta

    m1 : _Meta

    no_overlap : bool, optional
        See the `aggregate` function for details.

    contiguous : bool, optional
        See the `aggregate` function for details.
   
    rtol : float, optional
        See the `aggregate` function for details.

    atol : float, optional
        See the `aggregate` function for details.
   
    messages : bool, optional
        See the `aggregate` function for details.
   
    no_strict_units : bool, optional
        See the `aggregate` function for details.

:Returns:

    out : dict or bool

'''
    axis = m0.a_identity

    out = {}
    # Check formula_terms fields
#    for name, transform0 in m0.formula_terms.iteritems():

    for name in m0.formula_terms:
        key, transform0 = m0.field.transforms(name, exact=True).popitem()

#        transform1 = m1.formula_terms[name]
        transform1 = m1.field.transform(name, exact=True)

#        new_transform = Transform(standard_name=transform0.name)
        new_transform = Transform(standard_name=name)
          
        for term, value1 in transform1.iteritems():
            if (term not in transform0 and
                not (value1.isscalar and value1.datum(0) == 0.0)):
                if messages:
                    print("Won't aggregate %s: Incompatible %s transform" %
                          (repr(m0.field), repr(name)))
                return
        #--- End: for

        for term, value0 in transform0.iteritems():

            if (term not in transform1 and
                not (value0.isscalar and value0.datum(0) == 0.0)):
                if messages:
                    print(
"Won't aggregate %s: Incompatible %s transform term: %s" %
(repr(m0.field), repr(name), repr(term)))
                return
            #--- End: if

            value1 = transform1[term]
  
            if value0 in transform0.coordinates and value1 in transform1.coordinates:
                # Both transform values are pointers to coordinates in
                # their respective fields
                new_transform[term] = value0
                continue
            #--- End: if

#            if (isinstance(value0, basestring) and
#                isinstance(value1, basestring)):
#                # ----------------------------------------------------
#                # Both transform values are pointers to coordinates in
#                # their respective fields
#                # ----------------------------------------------------
#                new_transform[term] = value0
#                continue
#            #--- End: if
           
#            if isinstance(value0, basestring) or isinstance(value1, basestring):
            if value0 in transform0.coordinates or value1 in transform1.coordinates:
                # One transform value is a pointer to a coordinate and
                # the other is not
                if messages:
                    print(
"Won't aggregate %s: Incompatible %s transform term: %s" %
(repr(m0.field), repr(name), repr(term)))
                return
            #--- End: if
                        
            # --------------------------------------------------------
            # Both transform values are fields
            # --------------------------------------------------------
            x0 = _Meta(value0, messages=messages,
                       no_strict_units=no_strict_units, no_strict_identity=True)
            x1 = _Meta(value1, messages=messages,
                       no_strict_units=no_strict_units, no_strict_identity=True)

            if not (x0 and x1):
                # At least one field doesn't have a structual
                # signature
                if messages:
                    print(
"Won't aggregate %s: Incompatible %s transform term: %s" %
(repr(m0.field), repr(name), repr(term)))
                return
            #--- End: if

            if axis not in x0.axis and axis not in x1.axis:
                # Neither field spans the aggregating axis ...
                if value0.equivalent_data(value1, rtol=rtol, atol=atol):
                    # ... and the fields have equivalent data
                    # arrays. Therefore we don't need to do any
                    # aggregation.
                    new_transform[term] = value0
                    continue
                else:
                    # ... and the fields do not have equivalent data
                    if messages:
                        print(
"Won't aggregate %s: Incompatible %s transform term: %s. Neither value spans the %s axis and their data are different" %
(repr(m0.field), repr(name), repr(term), repr(axis)))
                    return
            #--- End: if

            if not (axis in x0.axis and axis in x1.axis):
                # Only one of the fields spans the aggregating axis
                if messages:
                    print(
"Won't aggregate %s: Incompatible %s transform term: %s. Only one value spans the %s axis" %
(repr(m0.field), repr(name), repr(term), repr(axis)))
                return
            #--- End: if
            
            # Both fields span the aggregating axis, so try to
            # aggregate them.
            new_value = aggregate((value0, value1), messages=messages,
                                  no_overlap=no_overlap, contiguous=contiguous,
                                  no_strict_units=no_strict_units,
                                  no_strict_identity=True,
                                  axes=axis)

            if len(new_value) == 2:
                # Couldn't aggregate them (because we got two fields
                # back instead of one)
                if messages:
                    print(
"Won't aggregate %s: Incompatible %s transform term: %s" %
(repr(m0.field), repr(name), repr(term)))
                return
            #--- End: if

            # Successfully aggregated the transform fields
            transform0[term] = new_value[0]
            new_transform[term] = new_value[0]
        #---End: for

        out[key] = new_transform
    #---End: for

    return out
#--- End: def

def _aggregate_ancillary_variables(m0, m1, rtol=None, atol=None,
                                   no_strict_units=False, no_overlap=False,
                                   messages=False, contiguous=False):
    '''

Aggregate the ancillary variable fields.

:Parameters:

    m0 : _Meta

    m1 : _Meta

    no_overlap : bool, optional
        See the `aggregate` function for details.

    contiguous : bool, optional
        See the `aggregate` function for details.
   
    rtol : float, optional
        See the `aggregate` function for details.

    atol : float, optional
        See the `aggregate` function for details.
   
    messages : bool, optional
        See the `aggregate` function for details.
   
    no_strict_units : bool, optional
        See the `aggregate` function for details.

:Returns:

    out : cf.AncillaryVariables or bool

'''
    axis = m0.a_identity

    new_ancillary_variables = AncillaryVariables()

    # Check formula_terms fields
    for identity, field0 in m0.ancillary_variables.iteritems():
        field1 = m1.ancillary_variables[identity]

        x0 = _Meta(field0, messages=messages, no_strict_units=no_strict_units,
                   no_strict_identity=False)
        x1 = _Meta(field1, messages=messages, no_strict_units=no_strict_units,
                   no_strict_identity=False)
        
        if not (x0 and x1):
            # At least one field doesn't have a structual signature
            if messages:
                print(
                    "Won't aggregate %s: Incompatible %s ancillary variable" %
                    (repr(m0.field), repr(identity)))
            return
        #--- End: if
            
        if axis not in x0.axis and axis not in x1.axis:
            # Neither field spans the aggregating axis ...
            if field0.equivalent_data(field1, rtol=rtol, atol=atol, traceback=True):
                # ... and the fields are equivalent
                new_ancillary_variables.append(field0)
                continue

            # ... and the fields are not equivalent
            if messages:
                print(
"A matching pair of ancillary variables which don't span the aggregating axis have different values")
                print(
                    "Won't aggregate %s: Incompatible %s ancillary variable" %
                    (repr(m0.field), repr(identity)))
                return
        #--- End: if

        if not (axis in x0.axis and axis in x1.axis):
            if messages:
                print(
"Only one of a matching pair of ancillary variables spans the aggregating axis")
                print(
                    "Won't aggregate %s: Incompatible %s ancillary variable" %
                    (repr(m0.field), repr(identity)))
            return
        #--- End: if
            
        # Both fields span the aggregating axis, so try to aggregate
        # them.
        new_value = aggregate((field0, field1), messages=messages, 
                              no_overlap=no_overlap, contiguous=contiguous,
                              no_strict_units=no_strict_units,
                              no_strict_identity=True,
                              axes=axis)
        
        if len(new_value) == 2:
            # Couldn't aggregate them (because we got two fields back
            # instead of one)
            if messages:
                print(
                    "Won't aggregate %s: Incompatible %s ancillary variable" %
                    (repr(m0.field), repr(identity)))
            return
        #--- End: if

        m0.ancillary_variables[identity] = new_value[0]

        new_ancillary_variables.append(new_value[0])
    #---End: for

    return new_ancillary_variables
#--- End: def
