from numpy import array        as numpy_array
from numpy import column_stack as numpy_column_stack
from numpy import fromfile     as numpy_fromfile
from numpy import memmap       as numpy_memmap
from numpy import result_type  as numpy_result_type
from numpy import where        as numpy_where

from numpy.ma import array        as numpy_ma_array
from numpy.ma import column_stack as numpy_ma_column_stack
from numpy.ma import isMA         as numpy_ma_isMA 
from numpy.ma import masked_where as numpy_ma_masked_where
from numpy.ma import where        as numpy_ma_where

from unwgdos import unwgdos

from ..constants import _file_to_fh
from ..functions import (open_files_threshold_exceeded,
                         close_one_file,
                         parse_indices,
                         get_subspace)

from ..data.filearray import FileArray

_filename_to_file = _file_to_fh.setdefault('PP', {})

# ====================================================================
#
# PPFileArray object
#
# ====================================================================

class PPFileArray(FileArray):
    ''' 
    
A sub-array stored in a PP file.
    
**Initialization**

:Parameters:

    file : str
        The PP file name in normalized, absolute form.

    file_offset : int
        The start position in the file of the data array.

    dtype : numpy.dtype
        The numpy data type of the data array.

    ndim : int
        The number of dimensions in the data array.

    shape : tuple
        The data array's dimension sizes.

    size : int
        The number of elements in the data array.

    add_offset : 
         The numeric additive offset (in the CF sense) of the
         sub-array. A value of ``0`` means no additive offset.
  
    scale_factor :  
         The numeric scale factor (in the CF sense) of the sub-array.
         A value of ``1`` means no scale factor.

    lblrec : int
        The number of words that the sub-array occupies in the
        file. If the sub-array is unpacked then this will equal
        *size*, otherwise it will be les than *size*.

    lbpack : int
         The integer PP packing code of the sub-array. A value of
         ``0`` means that the sub-array is unpacked.

    binary_mask : bool

    _FillValue : optional 
        The missing data value for the sub-array.

**Examples**

>>> ppfile
<open file 'file.pp', mode 'rb' at 0xc45e00>
>>> a = PPFileArray(file=ppfile.name, file_offset=ppfile.tell(),
...                 dtype=numpy.dtype('float32'), shape=(73, 96), size=7008,
...                 ndim=2, lblrec=7008, lbpack=0, scale_factor=1,
...                 add_offset=0)

'''
    def __getitem__(self, indices):
        '''

x.__getitem__(indices) <==> x[indices]

Returns a numpy array.

''' 
        pp = self.open()

        lbpack = self.lbpack

        if not lbpack:
            # --------------------------------------------------------
            # Data array is not packed
            # --------------------------------------------------------            
            if indices is Ellipsis:
                delete_memmap = False
            
                pp.seek(self.file_offset, 0)
                array = numpy_fromfile(pp, dtype=self.dtype, count=self.lblrec)
                array.resize(self.shape)
            else:
                delete_memmap = True

                mm_array = numpy_memmap(pp, mode='r',
                                        offset=self.file_offset,
                                        dtype=self.dtype,
                                        shape=self.shape)
                
                indices = parse_indices(mm_array, indices)               
                array   = get_subspace(mm_array, indices)                
                array   = array.copy() ## ???
        else:
            # --------------------------------------------------------
            # Data array is packed
            # --------------------------------------------------------
            delete_memmap = False

            pp.seek(self.file_offset, 0)
            array = pp.read(self.lblrec*4)

            if lbpack == 1:
                # ----------------------------------------------------
                # WGDOS packing
                # ----------------------------------------------------
                array = unwgdos(array, 4, self.size, self._FillValue)
                array.resize(self.shape)

            else:
                raise ValueError(
                    "PP data array has unknown packing flag: lbpack=%d" % 
                    lbpack)
            
            if indices is not Ellipsis:
                indices = parse_indices(array, indices)               
                array   = get_subspace(array, indices)
        #--- End: if

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        fill_value = self._FillValue
        if fill_value is not None:
            # A fill value is set so mask any missing values
            mask = (array == fill_value)
            masked = mask.any()
            if masked:
                array = numpy_ma_masked_where(mask, array, copy=False)              
#                array = numpy_ma_array(array, mask=mask, fill_value=fill_value,
#                                       copy=False)
        else:
            # _FillValue is not set, so there are no missing values
            array = numpy_array(array, copy=True) ##???
            masked = False

        if delete_memmap:
            # Close the memory mapped file
            del mm_array

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        if self.scale_factor != 1:
            array *= self.scale_factor

        if self.add_offset:
            array += self.add_offset

        if self.binary_mask:
            if not masked:
                array = numpy_where(array!=0, 1, 0)
            else:
                array = numpy_ma_where(array!=0, 1, 0) ## dch
        #--- End: if

        # ------------------------------------------------------------
        # Return the numpy array
        # ------------------------------------------------------------
        return array
    #--- End: def

    def __repr__(self):
        '''

x.__repr__() <==> repr(x)

'''
        return '<CF %s: %s>' % (self.__class__.__name__, self)
    #--- End: def
   
    def __str__(self):
        '''

x.__str__() <==> str(x)

'''
        return '%s offset=%s' % (self.file, self.file_offset)
    #--- End: def

    def close(self):
        '''

Close the file containing the data array.

If the file is not open then no action is taken.

:Returns:

    None

**Examples**

>>> f.close()

'''
        pp = _filename_to_file.pop(self.file, None)
        if pp is not None:
            pp.close()
    #--- End: def
   
    def open(self):
        '''

Return the 

:Returns:

    out : file

**Examples**

>>> f.open()

'''    
        filename = self.file

        pp = _filename_to_file.get(filename, None)
        
        if pp is not None:
            # File is already open
            return pp
      
        if open_files_threshold_exceeded():
            # Close an arbitrary data file to make way for this one
            close_one_file()
            
        pp = open(filename, 'rb')

        # Update _file_to_Dataset dictionary
        _filename_to_file[filename] = pp

        return pp
    #--- End: def

#--- End: class


# ====================================================================
#
# PPFileArrayBounds object
#
# ====================================================================

class PPFileArrayBounds(FileArray):
    '''  
    
'''

    def __init__(self, **kwargs):
        '''

**Initialization**

:Parameters:

    _lower : PPFileArray

    _upper : PPFileArray

'''
        super(PPFileArrayBounds, self).__init__(**kwargs)

        lower = self._lower

        self.dtype = numpy_result_type(lower.dtype, self._upper.dtype)
        self.shape = lower.shape + (2,)       
        self.size  = lower.size * 2
        self.ndim  = lower.ndim + 1
    #--- End: def
   
    def __getitem__(self, indices):
        '''

x.__getitem__(indices) <==> x[indices]

Returns a numpy array.

'''
        # ------------------------------------------------------------
        # Read the upper and lower bounds from the PP file and stick
        # them together
        # ------------------------------------------------------------
        lower = self._lower[...]
        upper = self._upper[...]

        if numpy_ma_isMA(lower) or numpy_ma_isMA(upper):
            array = numpy_ma_column_stack((lower, upper))
        else:
            array = numpy_column_stack((lower, upper))

        indices = parse_indices(array, indices)

        return get_subspace(array, indices)
    #--- End: def
  
    def __str__(self):
        '''

x.__str__() <==> str(x)

'''
        return '%s%s' % (self.__class__.__name__, self.shape)
    #--- End: def
   
    def close(self):
        '''

Close the file containing the data array.

If the file is not open then no action is taken.

:Returns:

    None

**Examples**

>>> f.close()

'''
        # An instance references no open files
        pass
    #--- End: def
   
    def copy(self):
        '''

Return a deep copy.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

:Returns:

    out :
        A deep copy.
    
**Examples**

>>> g = f.copy()

'''  
        return type(self)(_lower=self._lower.copy(), _upper=self._upper.copy())
    #--- End: def
    
#--- End: class
